%array
%s spell fast

%{

/*
speak - speech synthesizer using digitized sounds
Author: Klaus Wachter
e-mail: mail@klaus-wachter-1.de
Copyright: Freeware
Platform: LINUX
*/

#include <string.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include "germphn.h"

#define YY_READ_BUF_SIZE 1

#ifdef CONNECTED
#define YY_INPUT(buf,result,max_size) \
{ \
int c = getchar(); \
result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
putchar(c); \
}
#endif

FILE *audio, *record;
int audio_on;
int samplesize = 8, samplerate = 8000, stereo = 0;

int look = 0;

void *soundtable = NULL;
#define MAX_SOUNDS 100
void *sound[MAX_SOUNDS];
int sounds;
size_t soundtable_size;

#define P(i) play_sound(i)
#define U(s) unputs(s);

#define SYNC_AUDIO fflush(audio); ioctl(fileno(audio), SNDCTL_DSP_SYNC);

%}

%%

	#ifdef CONNECTED
	setbuf(stdout, NULL);
	#endif

	BEGIN(fast);

\.P[0-9]+\x0a	{
	int sound_number;
	sscanf(yytext + 2, "%d", &sound_number);
	P(sound_number);
	}

\.rate[0-9]+\x0a	{
	sscanf(yytext + 5, "%d", &samplerate);
	ioctl(fileno(audio), SNDCTL_DSP_SPEED, &samplerate);
	ioctl(fileno(audio), SNDCTL_DSP_SYNC);
	}

\.on\/off\x0a	{
	#ifndef DEBUG
	if(audio_on)
	{
	fclose(audio);
	audio_on = 0;
	}
	else 
	{
	set_audio();
	audio_on = 1;
	}
	#endif
	}

\.sync\x0a	{
	#ifndef DEBUG
	SYNC_AUDIO;
	#ifdef RECORD_FILE
	fflush(record);
	#endif
	#endif
	}

\.reset\x0a	{
	#ifndef DEBUG
	fflush(audio);
	ioctl(fileno(audio), SNDCTL_DSP_RESET);
	#endif
	}

\.normal\x0a	BEGIN(0);

\.spell\x0a	BEGIN(spell);

\.fast\x0a	BEGIN(fast);

<spell>[\x80-\xff]	{
	unput(yytext[0] & 0x7f);
	unput(0x0a); U(".P26");
	}

<spell>\x00	U("nul");

<spell>\x01	U("soh");

<spell>\x02	U("stx");

<spell>\x03	U("etx");

<spell>\x04	U("eot");

<spell>\x05	U("enq");

<spell>\x06	U("ack");

<spell>\x07	U("bel");

<spell>\x08	U("bs");

<spell>\x09	U("ht");

<spell>\x0a	{
	P(Pae); P(Pl); P(Pae); P(Pf);
	#ifdef TERMINAL
	SYNC_AUDIO;
	#endif
	}

\x0a	{
	#ifdef TERMINAL
	SYNC_AUDIO;
	#endif
	}

<spell>\x0b	U("vt");

<spell>\x0c	U("ff");

<spell>\x0d	U("cr");

<spell>\x0e	U("so");

<spell>\x0f	U("si");

<spell>\x10	U("dle");

<spell>\x11	U("dc1");

<spell>\x12	U("dc2");

<spell>\x13	U("dc3");

<spell>\x14	U("dc4");

<spell>\x15	U("nak");

<spell>\x16	U("syn");

<spell>\x17	U("etb");

<spell>\x18	U("can");

<spell>\x19	U("em");

<spell>\x1a	U("sub");

<spell>\x1b	U("esc");

<spell>\x1c	U("fs");

<spell>\x1d	U("gs");

<spell>\x1e	U("rs");

<spell>\x1f	U("us");

[\x00-\x1f]	;

<spell>\x20+	P(Pl); P(Pae); P(Pr);

\x20+	P(Ppause);

\x21	P(Pr); P(Pu); P(Pf);

\x22	P(Pf); P(Pue); P(Pr);

\x23	P(Pn); P(Pu); P(Pm);

\x24	{
	if(look)
	{
	char c;
	c = input(); unput(c);
	if('a' <= c && c <= 'z') {P(Pk); P(Pl); P(Pa); P(Pi); P(Pn);}
	look = 0;
	}
	else {P(Pd); P(Po2); P(Pl);}
	}

\x25	P(Pp); P(Pr); P(Po);

\x26	P(Pu); P(Pn); P(Pd);

\x27	P(Pa); P(Pp); P(Po);

\x28	P(Pr); P(Pu); P(Pn); P(Pd); P(Pa); P(Pu); P(Pf);

\x29	P(Pr); P(Pu); P(Pn); P(Pd); P(Pt); P(Ps); P(Pu);

\x2a	P(Psch); P(Pt); P(Pae); P(Pr); P(Pn);

\x2b	P(Pp); P(Pl); P(Pu); P(Ps);

<fast>\x2c[\x0a\x20]	P(Ppause); P(Ppause);

\x2c	P(Pk); P(Po2); P(Pm); P(Pa);

\x2d	P(Psch); P(Pt); P(Pr); P(Pi); P(Pch);

<fast>\x2e[\x0a\x20]	P(Ppause); P(Ppause); P(Ppause);

\x2e	P(Pp); P(Pu); P(Png); P(Pk); P(Pt);

\x2f	P(Psch); P(Pr); P(Pae); P(Pg); P(Pa); P(Pu); P(Pf);

\x30	P(Pn); P(Pu); P(Pl);

\x31	P(Pa); P(Pe); P(Pn); P(Ps);

\x32	P(Pt); P(Ps); P(Pf2); P(Pa); P(Pe);

\x33	P(Pd); P(Pr); P(Pa); P(Pi);

\x34	P(Pf); P(Pi); P(Pr);

\x35	P(Pf); P(Pue); P(Pn); P(Pf);

\x36	P(Ps2); P(Pae); P(Pk); P(Ps);

\x37	P(Ps2); P(Pi); P(Pb); P(Pe); P(Pn);

\x38	P(Pa); P(Pch2); P(Pt);

\x39	P(Pn); P(Po2); P(Pi); P(Pn);

\x3a	P(Pd); P(Po2); P(Pp); P(Pl);

\x3b	P(Ps2); P(Pe); P(Pm); P(Pi);

\x3c	P(Psch); P(Pp); P(Pi); P(Pt); P(Ps); P(Pa); P(Pu); P(Pf);

\x3d	P(Pg); P(Pl); P(Pa); P(Pi); P(Pch);

\x3e	P(Psch); P(Pp); P(Pi); P(Pt); P(Ps); P(Pt); P(Ps); P(Pu);

\x3f	P(Pf); P(Pr); P(Pa); P(Pg);

\x40	P(Pae); P(Pt);

[A-Z\xc4\xd6\xdc]+	{
	int i;
	if(YY_START != fast)
	{P(Pg); P(Pr); P(Po); P(Ps);}
	if(yyleng == 1) unput(yytext[0] | 0x20);
	else {
	if(YY_START != fast)
	{P(Pg); P(Pr); P(Po); P(Ps);}
	unput('$');
	for(i = yyleng - 1; i >= 0; i--) unput(yytext[i] | 0x20);
	look = 1;
	}
	}

\x5b	P(Pae); P(Pk); P(Pa); P(Pu); P(Pf);

\x5c	P(Psch); P(Pr); P(Pae); P(Pg); P(Pa); P(Pb);

\x5d	P(Pae); P(Pk); P(Pt); P(Ps); P(Pu);

\x5e	P(Pd); P(Pa); P(Pch2);

\x5f	P(Pu); P(Pn); P(Pt); P(Pae); P(Pr);

\x60	P(Pg); P(Pr); P(Pa); P(Pf);

<fast>ach	P(Pa); P(Pch2);

<fast>ae	unput(0xe4);

<fast>\xe4u	P(Po2); P(Pi);

a	P(Pa);

<spell>b	P(Pb); P(Pe);

b	P(Pb);

<spell>c	P(Pt); P(Ps); P(Pe);

<fast>ch	P(Pch);

<fast>ck	P(Pk);

<fast>c	P(Pk);

c	P(Pch);

<spell>d	P(Pd); P(Pe);

d	P(Pd);

<fast>e/[lrx]	P(Pae);

<fast>ei	P(Pa); P(Pi);

<fast>eu	P(Po2); P(Pi);

e	P(Pe);

<spell>f	P(Pae); P(Pf);

<fast>ff	P(Pf);

f	P(Pf);

<spell>g	P(Pg); P(Pe);

g	P(Pg);

<spell>h	P(Ph); P(Pa);

h	P(Ph);

<fast>ie	P(Pi);

i	P(Pi);

<spell>j	P(Pi2); P(Po2); P(Pt);

j	P(Pi2);

<spell>k	P(Pk); P(Pa);

k	P(Pk);

<spell>l	P(Pae); P(Pl);

<fast>ll	P(Pl);

l	P(Pl);

<spell>m	P(Pae); P(Pm);

<fast>mm	P(Pm);

m	P(Pm);

<spell>n	P(Pae); P(Pn);

<fast>ng	P(Png);

<fast>nn	P(Pn);

n	P(Pn);

<fast>och	P(Po2); P(Pch2);

<fast>oe	unput(0xf6);

<fast>or	P(Po2); P(Pr);

o	P(Po);

<spell>p	P(Pp); P(Pe);

<fast>ph	P(Pf);

<fast>pp	P(Pp);

p	P(Pp);

<spell>q	P(Pk); P(Pu);

<fast>qu	P(Pk); P(Pf2);

q	P(Pch2);

<spell>r	P(Pae); P(Pr);

<fast>rr	P(Pr);

r	P(Pr);

<spell>s	P(Pae); P(Ps);

<fast>sch	P(Psch);

<fast>s/p	P(Psch); 

<fast>ss	P(Ps);

s	P(Ps);

<spell>t	P(Pt); P(Pe);

<fast>t/io	P(Pt); P(Ps);

<fast>tt	P(Pt);

<fast>tz	P(Pt); P(Ps);

t	P(Pt);

<fast>uch	P(Pu); P(Pch2);

<fast>ue	unput(0xfc);

u	P(Pu);

<spell>v	P(Pf); P(Pa); P(Po2);

<fast>v	P(Pf);

v	P(Pf2);

<spell>w	P(Pf2); P(Pe);

w	P(Pf2);

<spell>x	P(Pi); P(Pk); P(Ps);

x	P(Pk); P(Ps);

<spell>y	P(Pue); P(Pp); P(Ps); P(Pi); P(Pl); P(Po2); P(Pn);

y	P(Pue);

<spell>z	P(Pt); P(Ps); P(Pae); P(Pt);

z	P(Pt); P(Ps);

\x7b	P(Psch); P(Pf2); P(Pa); P(Pae); P(Pf); P(Pa); P(Pu); P(Pf);

\x7c	P(Ps2); P(Pae); P(Png); P(Pk);

\x7d	P(Psch); P(Pf2); P(Pa); P(Pae); P(Pf); P(Pt); P(Ps); P(Pu);

\x7e	P(Psch); P(Pl); P(Pa); P(Png); P(Pae);

\x7f	U("del");

\x81	unput(0xfc);
\x84	unput(0xe4);
\x8e	unput(0xc4);
\x94	unput(0xf6);
\x99	unput(0xd6);
\x9a	unput(0xdc);
\xe1	unput(0xdf);

\xdf	P(Ps);

\xe4	P(Pae);

\xf6	P(Poe);

\xfc	P(Pue);

%%

main(argc, argv)
int argc;
char **argv;
{

#ifndef DEBUG

if(argc != 2) {fprintf(stderr, "Usage: %s <sound table>\n", argv[0]); exit(0);}

load_soundtable(argv[1]);

set_audio();

#else

sounds = MAX_SOUNDS;

#endif

yylex();

}

set_audio()
{
if((audio = fopen("/dev/dsp", "w")) != NULL)
{
audio_on = 1;
ioctl(fileno(audio), SNDCTL_DSP_SAMPLESIZE, &samplesize);
ioctl(fileno(audio), SNDCTL_DSP_STEREO, &stereo);
ioctl(fileno(audio), SNDCTL_DSP_SPEED, &samplerate);
}
else {
perror("/dev/dsp");
exit(1);
}
#ifdef RECORD_FILE
if((record = fopen(RECORD_FILE, "a")) == NULL)
perror(RECORD_FILE);
#endif
}

load_soundtable(filename)
char *filename;
{
int fd, i;
void *ptr;
int *intptr;
if((fd = open(filename, O_RDONLY)) == -1) {
perror(filename); exit(1);
}
else {
soundtable_size = (size_t)(lseek(fd, 0, SEEK_END));
lseek(fd, 0, SEEK_SET);
soundtable = realloc(soundtable, soundtable_size);
read(fd, soundtable, soundtable_size);
for(i = 0, ptr = soundtable;
ptr < soundtable + soundtable_size && i < MAX_SOUNDS;) {
sound[i] = ptr;
intptr = ptr;
if(*intptr < 0) {fprintf(stderr, "Incorrect sound table\n"); exit(1);}
/*
fprintf(stderr, "%d %d\n", (int)ptr, *intptr);
*/
ptr += *intptr + 4;
i++;
}
if(ptr > soundtable + soundtable_size) {fprintf(stderr, "Incorrect \
sound table\n"); exit(1);}
else sounds = i;
}
/*
fprintf(stderr, "sounds = %d, soundtable_size = %d\n", sounds,
(int)soundtable_size);
*/
}

play_sound(num)
int num;
{
int *intptr;
if(0 <= num && num < sounds && sounds > 0) {
intptr = sound[num];
#ifndef DEBUG
fwrite(sound[num] + 4, *intptr, 1, audio);
#ifdef RECORD_FILE
fwrite(sound[num] + 4, *intptr, 1, record);
#endif
#else
fprintf(stderr, "p%i ", num);
#endif
}
}

unputs(s)
char *s;
{
int i;
for(i = strlen(s) - 1; i >= 0; i--) unput(s[i]);
}
