#include <string.h>
#include <fcntl.h>
#include <gpm.h>

#include "germphn.h"

#ifdef CONNECTED
int datapipe_fd[2], handshakepipe_fd[2];
pid_t speak_pid;
#endif

#define P(n) put_sound(n)

int vc;
struct
{
  char lines, columns, x, y;
}
screen;
int right = 0, bottom = 0;

typedef struct
{
  char x, y;
}
position;
position pos;

int asr_mode;
#define PAGE 1
#define LINE 2

char sound_command[10];

int speak_mode;
#define SPELL 1
#define NORMAL 2
#define FAST 3

#define DELAY 200
int long_down = 0, skip_events = 0;

#include "actions.c"

Gpm_Event event;

mev_handler ()
{

  if (skip_events)
    {
      skip_events--;
      return;
    }
  if (long_down)
    {
      long_down = 0;
      return;
    }

  if (event.type == GPM_MOVE)
    {
      move ();
      RET;
    }


  if (Gpm_Repeat (DELAY))
    {

      if ((event.type == (GPM_UP | GPM_SINGLE)))
	{
	  if (event.buttons == GPM_B_LEFT)
	    {
	      left_1 ();
	      RET;
	    }
	  if (event.buttons == GPM_B_MIDDLE)
	    {
	      middle_1 ();
	      RET;
	    }
	  if (event.buttons == GPM_B_RIGHT)
	    {
	      right_1 ();
	      RET;
	    }
	}

      if ((event.type == (GPM_UP | GPM_DOUBLE)))
	{
	  if (event.buttons == GPM_B_LEFT)
	    {
	      left_2 ();
	      RET;
	    }
	  if (event.buttons == GPM_B_MIDDLE)
	    {
	      middle_2 ();
	      RET;
	    }
	  if (event.buttons == GPM_B_RIGHT)
	    {
	      right_2 ();
	      RET;
	    }
	}

      if ((event.type == (GPM_DOWN | GPM_SINGLE)))
	{
	  if (event.buttons == GPM_B_LEFT)
	    {
	      left_long ();
	      RET;
	    }
	  if (event.buttons == GPM_B_MIDDLE)
	    {
	      middle_long ();
	      RET;
	    }
	  if (event.buttons == GPM_B_RIGHT)
	    {
	      right_long ();
	      RET;
	    }
	}

    }

}

#ifdef CONNECTED

putchar_wait (c)
     char c;
{
  putchar (c);
  getchar ();
}


putstring_wait (s)
     char *s;
{
  int i, l;
  l = strlen (s);
  for (i = 0; i < l; i++)
    putchar_wait (s[i]);
}

#else

#define putchar_wait(c) putchar(c)

#define putstring_wait(s) fputs(s, stdout)

#endif

put_sound (int sound_number)
{
  sprintf (sound_command, ".P%i\n", sound_number);
  putstring_wait (sound_command);
}

main (argc, argv)
     int argc;
     char **argv;
{

  Gpm_Connect conn;

#ifdef CONNECTED
  if (argc != 3)
    {
      fprintf (stderr, "Usage: %s <synthesizer> <sound table>\n", argv[0]);
      exit (0);
    }
  pipe (datapipe_fd);
  pipe (handshakepipe_fd);
  speak_pid = fork ();
  if (speak_pid == -1)
    {
      perror ("fork");
      exit (1);
    }
  if (speak_pid == 0)
    {
      dup2 (datapipe_fd[0], 0);
      close (datapipe_fd[1]);
      dup2 (handshakepipe_fd[1], 1);
      close (handshakepipe_fd[0]);
      execlp (argv[1], "speak", argv[2], NULL);
    }
  else
    {
      dup2 (datapipe_fd[1], 1);
      close (datapipe_fd[0]);
      dup2 (handshakepipe_fd[0], 0);
      close (handshakepipe_fd[1]);
      setbuf (stdout, NULL);
    }
#endif

  asr_mode = PAGE;
  speak_mode = FAST;
  set_speak ();

  /* build a catch-all connectionData */
  conn.eventMask = ~0;
  conn.defaultMask = 0;
  conn.maxMod = ~0;
  conn.minMod = 0;
  if (Gpm_Open (&conn, -1) == -1)
    {
      perror ("Gpm_Open");
      exit (1);
    }

  if ((vc = open ("/dev/vcsa0", O_RDONLY)) == -1)
    {
      perror ("Vc");
      exit (1);
    }
  read (vc, &screen, 4);
  right = screen.columns - 1;
  bottom = screen.lines - 1;
  daemon (1, 1);
  while (1)
    {
      Gpm_GetEvent (&event);
      mev_handler ();
    }
  Gpm_Close ();
  exit (0);
}

unblock (fd)
     int fd;
{
  int flag;

  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag | O_NONBLOCK);
}

block (fd)
     int fd;
{
  int flag;

  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag & ~(O_NONBLOCK));
}

get_char (position pos)
{
  int offset;
  char c;
  offset = 4 + 2 * ((pos.y * screen.columns) + pos.x);
  lseek (vc, offset, SEEK_SET);
  read (vc, &c, 1);
  return (c);
}

output_char (position pos)
{
  putchar_wait (get_char (pos));
}

output_line_segment (int dx)
{
  int direction, end;
  position pos_temp;
  if (dx == 0)
    {
      output_char (pos);
      return;
    }
  if (dx > 0)
    direction = 1;
  if (dx < 0)
    direction = -1;
  end = pos.x + dx;
  if (end < 0)
    end = 0;
  if (end > right)
    end = right;
  pos_temp = pos;
  if (direction == 1)
    {
      while (pos_temp.x <= end)
	{
	  output_char (pos_temp);
	  pos_temp.x++;
	}
    }
  if (direction == -1)
    {
      while (pos_temp.x >= end)
	{
	  P (Pklick);
	  pos_temp.x--;
	}
    }
}

next_position ()
{
  if ((pos.x == right) && (pos.y == bottom))
    return 0;
  if (pos.x < right)
    pos.x++;
  else
    {
      pos.x = 0;
      pos.y++;
    }
  return 1;
}

output_screen ()
{
  unblock (gpm_fd);
  output_char (pos);
  while (next_position ())
    {
      /*
         if (Gpm_GetSnapshot (&event) != 0)
       */
      if (Gpm_GetEvent (&event) == -1)
	output_char (pos);
      else
	{
	  if (event.buttons == GPM_B_LEFT)
	    {
	      back ();
	      pos.x = -1;
	      reset_audio ();
	      P (Ppause);
	      P (Pklick);
	      P (Ppause);
	    }
	  else
	    if ((event.buttons == GPM_B_MIDDLE)
		&& (event.type == (GPM_DOWN | GPM_SINGLE)))
	    {
	      skip_events = 1;
	      reset_audio ();
	      block (gpm_fd);
	      return;
	    }
	  else if (event.buttons == GPM_B_RIGHT)
	    {
	      down_one ();
	      pos.x = -1;
	      reset_audio ();
	      P (Ppause);
	      P (Pklick);
	      P (Ppause);
	    }
	  else
	    output_char (pos);
	}
    }
  pos.x = 0;
  pos.y = bottom;
  block (gpm_fd);
}

set_speak ()
{
  if (speak_mode == SPELL)
    putstring_wait (".spell\n");
  if (speak_mode == NORMAL)
    putstring_wait (".normal\n");
  if (speak_mode == FAST)
    putstring_wait (".fast\n");
}

change_line ()
{
  int i, temp;
  temp = pos.y;
  pos.y += event.dy;
  if (pos.y < 0)
    pos.y = 0;
  if (pos.y > bottom)
    pos.y = bottom;
  temp = abs (pos.y - temp);
  if (temp > 0)
    {
      pos.x = 0;
      for (i = 1; i <= temp; i++)
	{
	  P (Pklick);
	}
    }
}

change_column ()
{
  int temp;
  temp = pos.x;
  temp += event.dx;
  if (temp < 0)
    temp = 0;
  if (temp > right)
    temp = right;
  temp = temp - pos.x;
  if (temp > 0)
    {
      pos.x++;
      output_line_segment (temp - 1);
      pos.x += temp - 1;
    }
  if (temp < 0)
    {
      pos.x--;
      output_line_segment (temp + 1);
      pos.x += temp + 1;
    }
}

up_one ()
{
  pos.y--;
  if (pos.y < 0)
    pos.y = 0;
}

back ()
{
  if (pos.x > 0)
    pos.x = 0;
  else
    up_one ();
}

left_one ()
{
  if (pos.x == 0)
    return;
  pos.x--;
  output_char (pos);
}

goto_first ()
{
  pos.x = 0;
  pos.y = 0;
}

output_line ()
{
  pos.x = 0;
  while (pos.x <= right)
    {
      if (Gpm_GetSnapshot (&event) != 0)
	output_char (pos);
      else
	{
	  skip_events = 3;
	  reset_audio ();
	  return;
	}
      pos.x++;
    }
  pos.x = 0;
  down_one ();
}

goto_cursor ()
{
  lseek (vc, 0, SEEK_SET);
  read (vc, &screen, 4);
  /*
     fprintf(stderr, "%d %d %d %d\n", 
     screen.lines, screen.columns, screen.x, screen.y);
   */
  pos.x = screen.x;
  pos.y = screen.y;
}

goto_end ()
{
  pos.x = right;
  while (((get_char (pos) == 0x20)) && (pos.x > 0))
    pos.x--;
}

switch_audio ()
{
  P (Po2);
  P (Pf);

  putstring_wait (".on/off\n");
  P (Po2);
  P (Pn);
  sync_audio ();
}

switch_speak_mode ()
{
  if (speak_mode == SPELL)
    {
      speak_mode = NORMAL;
      set_speak ();
      P (Pn);
      P (Po2);
      P (Pr);
      P (Pm);
      P (Pa);
      P (Pl);
    }
  else if (speak_mode == NORMAL)
    {
      speak_mode = FAST;
      set_speak ();
      P (Pf);
      P (Pa);
      P (Ps);
      P (Pt);
    }
  else if (speak_mode == FAST)
    {
      speak_mode = SPELL;
      set_speak ();
      P (Ps);
      P (Pp);
      P (Pae);
      P (Pl);
    }
  sync_audio ();
}

switch_asr_mode ()
{
  if (asr_mode == PAGE)
    {
      asr_mode = LINE;
      putstring_wait (".spell\n");
      P (Pl);
      P (Pa);
      P (Pe);
      P (Pn);
    }
  else if (asr_mode == LINE)
    {
      asr_mode = PAGE;
      set_speak ();
      P (Pp);
      P (Pe);
      P (Pi);
      P (Pd);
      P (Psch);
    }
  sync_audio ();
}

down_one ()
{
  pos.y++;
  if (pos.y > bottom)
    pos.y--;
}

right_one ()
{
  if (pos.x == right)
    return;
  pos.x++;
  output_char (pos);
}

goto_last ()
{
  pos.x = 0;
  pos.y = bottom;
}

sync_audio ()
{
  putstring_wait (".sync\n");
}

reset_audio ()
{
  putstring_wait (".reset\n");
}

print_event ()
{
  printf ("but=x%x mod=x%x vc=%i \
x=%2i y=%2i dx=%3i dy=%3i type=x%2x clicks=%i marg=x%x\n", event.buttons, event.modifiers, event.vc, event.x, event.y, event.dx, event.dy, event.type, event.clicks, event.margin);
}
