%{

/*
ste - sound table editor
Author: Klaus Wachter
e-mail: mail@klaus-wachter-1.de
Copyright: Freeware
Platform: LINUX
*/

#include <fcntl.h>
#include <termios.h>
#include <sys/soundcard.h>

#define YY_READ_BUF_SIZE 1

struct termios terminal_flags_orig, terminal_flags;

int num;
char *nam = NULL;
int audio;
int sample_size = 8, sample_rate = 8000, stereo = 0;
char *buffer = NULL;
int buffer_size = 4096;
int window_size, sounds_per_second = 10;
int zero_level = 0x80808080;
int frame_size;
int step;
int low, high; 

void *soundtable = NULL;
#define MAX_SOUNDS 100
void *sound[MAX_SOUNDS];
int sounds;
size_t soundtable_size;
int mult = 3;

%}

%%

	tcgetattr(0, &terminal_flags_orig);
	tcgetattr(0, &terminal_flags);
	(&terminal_flags)->c_lflag &= ~(ECHO|ICANON);
	tcsetattr(0, TCSANOW, &terminal_flags);

	#ifdef VERIFY_AUDIO
	set_audio(O_RDONLY); close(audio);
	#endif

	buffer = realloc(buffer, buffer_size);

	set_zero();
	set_frame_size();
	set_step();
	set_window_size();
	
	printf("Type \"?\" for help.\n");


\?	help();

[0-9]+	num = atoi(yytext); printf("num = %d\n", num);

\+	num++; printf("num = %d\n", num);

\-	if(num > 0) num--; printf("num = %d\n", num);

\"[^\"]*\"	{
	nam = realloc(nam, yyleng - 1);
	yytext[yyleng - 1] = 0;
	strcpy(nam, yytext + 1);
	printf("nam = \"%s\"\n", nam);
	}

sr	{
	sample_rate = num;
	set_audio(O_RDONLY); close(audio);
	set_window_size();
	printf("sample_rate = %d\n", sample_rate);
	}

ss	{
	if(sample_size == 8) sample_size = 16; else sample_size = 8;
	set_zero();
	set_frame_size();
	printf("sample_size = %d\n", sample_size);
	}

st	{
	if(stereo == 0) stereo = 1; else stereo = 0;
	set_frame_size();
	printf("stereo = %d\n", stereo);
	}

b	if(num >= 1) {
	buffer_size = num + (frame_size - ((num - 1) % frame_size)) - 1; 
	buffer = realloc(buffer, buffer_size);
	set_step();
	printf("buffer_size = %d, step = %d\n", buffer_size, step);
	}

a	low -= frame_size; if(low < 0) low = 0;
f	low += frame_size; if(low > high) low = high;
j	high -= frame_size; if(high < low) high = low;
;	high += frame_size; if(high > buffer_size) high = buffer_size;
A	low -= step; if(low < 0) low = 0; play_buf_mult();
F	low += step; if(low > high) low = high; play_buf_mult();
J	high -= step; if(high < low) high = low; play_buf_mult();
:	high += step; if(high > buffer_size) high = buffer_size; play_buf_mult();

c	{
	low -= frame_size; if(low < 0) low = 0;
	high = low + (window_size * frame_size);
	if(high > buffer_size) high = buffer_size;
	}

,	{
	low += frame_size; if(low > buffer_size) low = buffer_size;
	high = low + (window_size * frame_size);
	if(high > buffer_size) high = buffer_size;
	}

w	{
	if(num > 0) window_size = num;
	printf("window_size = %d\n", window_size);
	}

W	{
	if(num > 0) sounds_per_second = num;
	set_window_size();
	printf("window_size := %d sounds / second\n", sounds_per_second);
	}

p	play_buf();

P	play_buf_mult();

m	if(num > 0) mult = num; printf("mult = %d\n", mult);

r$	record_buf(); printf("new buffer recorded\n");

S$	append_sound();

l	load_soundtable();

t	play_sound(num);

T	{int i;
	if(sounds == 0) fprintf(stderr, "no sounds\n");
	else {
	printf("playing whole sound table\n");
	for(i = 0; i < sounds; i++) play_sound(i);
	}
	}

x	extract_sound();

I$	insert_sound();

D$	delete_sound();

C$	change_sound();

z	{
	printf("Entering feedback mode; stop with any key");
	feedback();
	printf("\n");
	}

q	quit(0);

\.	{
	printf("\
\n\
Settings:\n\
num = %d, nam = \"%s\"\n\
sample_rate = %d, sample_size = %d, stereo = %d\n\
buffer_size = %d, windos_size = %d, high - low = %d, mult = %d\n\
sounds = %d\n\
",
	num, nam,
	sample_rate, sample_size, stereo,
	buffer_size, window_size, high - low, mult,
	sounds);
	}
	
!.+$	printf("%s\n", yytext); system(yytext + 1);

%%

main (int ac, char **av)
{
  if (ac == 2)
    nam = av[1];
  if (ac > 2)
    {
      fprintf (stderr, "Usage: %s [<name>]\n", av[0]);
      exit (1);
    }
  yylex ();
}

unblock (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag | O_NONBLOCK);
}

block (fd)
     int fd;
{
  int flag;
  flag = fcntl (fd, F_GETFL);
  fcntl (fd, F_SETFL, flag & ~(O_NONBLOCK));
}

set_audio (o_flags)
     int o_flags;
{
  if ((audio = open ("/dev/dsp", o_flags, 0)) == -1)
    {
      perror ("/dev/dsp");
      quit (1);
    }
  ioctl (audio, SNDCTL_DSP_SAMPLESIZE, &sample_size);
  ioctl (audio, SNDCTL_DSP_STEREO, &stereo);
  ioctl (audio, SNDCTL_DSP_SPEED, &sample_rate);
}

/* unused */
play_file (start, end)
     off_t start, end;
{
  int fd, n_read;
  char c;
  if ((fd = open (nam, O_RDONLY)) == -1)
    perror (nam);
  else
    {
      lseek (fd, start, SEEK_SET);
      set_audio (O_WRONLY);
      unblock (0);
      while (
	     ((end == 0) || (lseek (fd, 0, SEEK_CUR) <= end))
	     && ((c = getchar ()) == -1)
	     && ((n_read = read (fd, buffer, buffer_size)) > 0))
	write (audio, buffer, n_read);
      block (0);
      unput (c);
      close (fd);
      close (audio);
    }
}

record_buf ()
{
  set_audio (O_RDWR);
  low = 0;
  high = buffer_size;
  read (audio, buffer, buffer_size);
  write (audio, buffer, buffer_size);
  close (audio);
}

play_zero ()
{
  write (audio, (char *) &zero_level, 4);
}

play_buf ()
{
  set_audio (O_WRONLY);
  play_zero ();
  write (audio, buffer + low, high - low);
  play_zero ();
  close (audio);
}

play_buf_mult ()
{
  int i;
  set_audio (O_WRONLY);
  for (i = 0; i < mult; i++)
    {
      play_zero ();
      write (audio, buffer + low, high - low);
      play_zero ();
    }
  close (audio);
}

/* unused */
play_buf_always ()
{
  char c;
  unblock (0);
  while ((c = getchar ()) == -1)
    play_buf ();
  block (0);
  unput (c);
}

append_sound ()
{
  int fd, sound_length;
  if ((fd = open (nam, O_WRONLY | O_CREAT | O_APPEND, 0666)) == -1)
    perror (nam);
  else
    {
      sound_length = high - low;
      write (fd, &sound_length, 4);
      write (fd, buffer + low, sound_length);
      close (fd);
      printf ("appended sound of length %d bytes to file \"%s\"\n",
	      sound_length, nam);
    }
}

load_soundtable ()
{
  int fd, i;
  void *ptr;
  int *intptr;
  if ((fd = open (nam, O_RDONLY)) == -1)
    perror (nam);
  else
    {
      soundtable_size = (size_t) (lseek (fd, 0, SEEK_END));
      lseek (fd, 0, SEEK_SET);
      soundtable = realloc (soundtable, soundtable_size);
      read (fd, soundtable, soundtable_size);
      sounds = 0;
      for (i = 0, ptr = soundtable;
	   ptr < soundtable + soundtable_size && i < MAX_SOUNDS;)
	{
	  sound[i] = ptr;
	  intptr = ptr;
	  if (*intptr < 0)
	    {
	      fprintf (stderr, "Incorrect sound table\n");
	      return;
	    }
/*
   fprintf(stderr, "%d %d\n", (int)ptr, *intptr);
 */
	  ptr += *intptr + 4;
	  i++;
	}
      if (ptr > soundtable + soundtable_size)
	fprintf (stderr, "Incorrect \
sound table\n");
      else
	{
	  sounds = i;
	  printf ("loaded sound table from file \"%s\" with %d sounds\n", nam,
		  sounds);
	}
    }
/*
   fprintf(stderr, "sounds = %d, soundtable_size = %d\n", sounds, (int)soundtable_size);
 */
}

play_sound (n)
     int n;
{
  int *intptr;
  if (n < sounds && sounds > 0)
    {
      set_audio (O_WRONLY);
      intptr = sound[n];
      write (audio, sound[n] + 4, *intptr);
/*
   fprintf(stderr, "%d %d\n", (int)(sound[n] + 4), *intptr);
 */
      close (audio);
      printf ("played sound %d\n", n);
    }
  else
    fprintf (stderr, "sound %d?\n", n);
}

extract_sound ()
{
  int *intptr;
  if (num < sounds && sounds > 0)
    {
      intptr = sound[num];
      if (*intptr <= buffer_size)
	{
	  memcpy (buffer, sound[num] + 4, (size_t) * intptr);
	  low = 0;
	  high = *intptr;
	  printf ("extracted sound %d to buffer\n", num);
	}
    }
}

insert_sound ()
{
  int fd, sound_length;
  if (
      (num < sounds)
      && (fd = open (nam, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
    perror (nam);
  else
    {
      write (fd, sound[0], (int) (sound[num] - sound[0]));
      sound_length = high - low;
      write (fd, &sound_length, 4);
      write (fd, buffer + low, sound_length);
      write (fd, sound[num], (int) (sound[0] + soundtable_size - sound[num]));
      close (fd);
      printf
	("inserted buffer as sound %d into loaded table and wrote the new \
table to file \"%s\"\n", num, nam);
      load_soundtable ();
    }
}

delete_sound ()
{
  int fd, sound_length;
  if (
      (num < sounds)
      && (fd = open (nam, O_WRONLY | O_CREAT | O_TRUNC, 0666)) == -1)
    perror (nam);
  else
    {
      write (fd, sound[0], (int) (sound[num] - sound[0]));
      if (num + 1 < sounds)
	write (fd, sound[num + 1],
	       (int) (sound[0] + soundtable_size - sound[num + 1]));
      close (fd);
      printf
	("deleted sound %d from loaded table and wrote the new table to file \
\"%s\"\n", num, nam);
      load_soundtable ();
    }
}

change_sound ()
{
  insert_sound ();
  num++;
  delete_sound ();
  num--;
  printf
    ("== replaced sound %d in loaded sound table by buffer and wrote the \
new table to file \"%s\"\n", num, nam);
}

set_zero ()
{
  if (sample_size == 8)
    zero_level = 0x80808080;
  else
    zero_level = 0x00000000;
}

set_frame_size ()
{
  frame_size = 1;
  if (sample_size == 16)
    frame_size *= 2;
  if (stereo == 1)
    frame_size *= 2;
}

set_step ()
{
  step = ((buffer_size / 160) + 1) * 4;
}

set_window_size ()
{
  window_size = ((sample_rate * frame_size) / sounds_per_second);
}

feedback ()
{
  char c;
  set_audio (O_RDWR);
  unblock (0);
  while ((c = getchar ()) == -1)
    {
      read (audio, buffer, buffer_size);
      write (audio, buffer, buffer_size);
    }
  block (0);
  close (audio);
  unput (c);
}

quit (n)
     int n;
{
  tcsetattr (0, TCSANOW, &terminal_flags_orig);
  printf ("quitting\n");
  exit (n);
}

help ()
{
  printf ("\
 ? -> display this help screen; z -> feedback mode; q -> quit this program\n\
 <number> -> num = <number>; + -> increment num; - -> decrement num\n\
 \"<string>\" -> nam = <string>; sr -> sample_rate = num\n\
 st -> toggle between stereo and mono; ss -> toggle sample_size\n\
 b -> adjust buffer_size and step according to num\n\
 w -> window_size = num; W -> window+size := num sounds / second\n\
 a -> decrement low by 1 frame; A -> decrement low by step frames\n\
 f -> increment low by 1 frame; F -> increment low by step frames\n\
 j -> decrement high by 1 frame; J -> decrement high by step frames\n\
 ; -> increment high by 1 frame; : -> increment high by step frames\n\
 c -> decrement low by 1 frame and set high according to window_size\n\
 , -> increment low by 1 frame and set high according to window_size\n\
 p -> play buffer; m -> mult = num; P -> play buffer mult times\n\
 r<enter> -> record buffer\n\
S<enter> -> append sound in buffer to file \"nam\"\n\
 l -> load sound table from file \"nam\" into memory\n\
 t -> play sound num; T -> play all sounds from sound table\n\
 x -> extract sound num into buffer\n\
 I<enter> -> insert sound at num and write table to file \"nam\"\n\
 D<enter> -> delete sound num and write table to file \"nam\"\n\
 C<enter> -> replace sound num with the one in buffer and write table ...\n\
 . -> display settings\n\
 !<string><enter> -> execute the command <string>\n\
");
}

